! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: emlmerge_mod
!
! Authors:  Brendan Perry - NASA/GSFC (complete rewrite) brendan.h.perry@nasa.gov
!           Steve Snowden - NASA/GSFC (original algorithm)
!
! Description: Extended Source Analysis Software
! SLS 4 March 2010
!
! Comments:
!
! $Id: eml_mod.f90,v 1.4 2009/05/26 12:53:39 aibarra Exp $
!
!  Merges emllist.fits source lists from emldetect
!
!  EMLMERGE creates a master list of sources detected from a list
!  of observations which have been processed to the point of a creating
!  an emldetect source list.  Sources are compared and those within 2"
!  of each other have the one with the lower flux thrown out.  It assumes
!  that the source detection has been in two bands, 0.4-1.25 keV and 
!  2.0-7.2 keV, which can be produced by cheese-bands in the default mode.
!  This task was originally called merge_source_list
!
!  INPUT:
!    dirfile     : Filename of directory list
!    maxlikelim  : Lower ML limit for source detections
!    outlist     : output eml srclist, defaults to emllist_merged.fits
!    withemlfill : Has emlfill been run?
!
!    Example: emlmerge dirfile=dir.txt emlout=mergedeml.fits maxlikelim=15.0
!
!  OUTPUT:
!    Merged emllist_merged.fits source list

module emlmerge_mod

  use dal
  use param 
  use errorHandling
  use gofcoords_mod ! from goflib, formerly in subsn (SS)
  use gofio_mod     ! from goflib, formerly in subsn (SS)
  use gofutils_mod     ! from goflib, formerly in subsn (SS)

  implicit none

  private

  character(len=*), parameter :: &
       dirfilepar = "dirfile", &     ! text directory list
       mlimpar    = "maxlikelim", &  ! lower ML limit for source detections
       outemlpar  = "emlout", &      ! output merged list filename
       filledpar  = "withemlfill"    ! has emlfill been run?

  public emlmerge

contains

  !---------------------------------------------------------------------
  subroutine emlmerge
    !---------------------------------------------------------------------

    type(DataSetT) :: inEML_S, outEML_S
    type(TableT)   :: inEML_T, outEML_T
    type(ColumnT)  :: mlidsrc_C, idinst_C, idband_C, scts_C, &
         detml_C, flux_C, fluxerr_C, rate_C, rateerr_C, ra_C, dec_C, &
         radecerr_C, distnn_C
    !
    character(len=256) :: dirs_F, outEML_F, currpath_F
    character(len=256), dimension(:), allocatable :: fullpath, emllist
    character(len=242) :: inDIRS_F, currdir, inEML_F
    character(len=60) :: COMM
    character(len=80) :: dir
    character(len=14) :: eml_F
    integer :: i, j, ios, lun, lunout, nRows, err, status
    integer :: blockn, blockl, hdutype, hist1(1000), hist2(1000), &
         iband, ibandt(100), icol, idec, idist, &
         iflux, ifluxerr, ii, iinst, instt(100), iml,  &
         ira, iradecerr, irate, irateerr, is, iscts, &
         isour, isrc, k, lundir, lunlist, m, mm, n
    integer, dimension(:), pointer :: mlidsrc_JP, idinst_JP, idband_JP
    logical :: withemlfill, warndie, exists
    real(kind=single) :: maxlim
    real(kind=single) :: dist(100000), distt(100), flux(100000), &
         fluxerr(100000), flux1(100000), fluxerr1(100000), &
         flux2(100000), fluxerr2(100000), fluxt(100), punc, &
         fluxerrt(100), rate(100000), rateerr(100000), &
         rate1(100000), rateerr1(100000), rate2(100000), &
         rateerr2(100000), ml(100000), mlt(100), &
         ratet(100), rateerrt(100), radecerr(100000), &
         radecerrt(100), scts(100000), sctst(100), &
         tnf1, tnf2, tnr1, tnr2, &
         tf1, tfe1, tf2, tfe2, tr1, tre1, tr2, tre2
    real(kind=single) :: decave, decmax, decmin, delra, deldec,  histmax, &
         histrange, maxlikelim, raave, ramax, ramin, rarange, scale
    real(kind=double) :: dang, dec(100000), dect(100), ra(100000), &
         rat(100), dlat1, dlat2, dlong1, dlong2
    real, dimension(:), pointer :: scts_RP, detml_RP, flux_RP, fluxerr_RP, &
         rate_RP, rateerr_RP, ra_RP, dec_RP, radecerr_RP, distnn_RP

    !
    ! get the parameters from the par file
    !
    dirs_F      = stringParameter(dirfilepar)
    maxlim      = realParameter(mlimpar)
    outEML_F    = stringParameter(outemlpar)
    withemlfill = booleanParameter(filledpar)
    !
    ! params not mandatory. if not given, it defaults to basic name.
    !
    if (withemlfill) then ! sas task emlfill has been run
       eml_F    = 'emllist_filled.fits'
       if (trim(adjustl(outEML_F)) .eq. 'default') then
          outEML_F = 'emllist_filled_merged.fits'
       endif
    else
       eml_F    = 'emllist.fits'
       if (trim(adjustl(outEML_F)) .eq. 'default') then
          outEML_F = 'emllist_merged.fits'
       endif
    endif
    inquire(FILE=trim(adjustl(outEML_F)), EXIST=exists)
    if (exists) then
       if (clobber()) then
          write(errstr,*)"Outfile ",trim(adjustl(outEML_F))," exists but clobber set, overwriting."
          call message(AppMsg,Always)
          call deleteDataSet(trim(adjustl(outEML_F)))
       else
          errstr = "Output file " // trim(adjustl(outEML_F))// " exists but noclobber set."
          call error("noClobber",trim(adjustl(errstr)))
       endif
    endif
    !
    write(errstr,*) "Directory list text file  = ",trim(adjustl(dirs_F))
    call message(AppMsg,Always)
    write(errstr,*) "Output merged source list = ",trim(adjustl(outEML_F))
    call message(AppMsg,Always)
    write(errstr,*) "lower ML limit for source detections = ",maxlim
    call message(AppMsg,Always)
    !
    lun = GetLun() ! in esas_mod in goflib
    open(unit=lun,form='formatted',status='old',file=trim(adjustl(dirs_F)),iostat=ios)
    if (ios .ne. 0) then
       call error ("badTextFile",&
            "Input directories text file does not exist or is unreadable")
    endif
    allocate(fullpath(2500)) ! for now
    fullpath = ''
    j = 0
    !
    ! Find all the emllists, sanity check, make final doable list.
    ! Max number of files set to 2500, with ~40 sources per file,
    ! Max number of sources is 100000. This can be changed ???!!!
    do while (ios .eq. 0)
       read(lun,'(A)',iostat=ios) currdir
       i = i + 1
       if (i .gt. 2500) then ! change this???!!!
          call warning('MaxDirs2500',&
               "Maximum directories is 2500, only doing first 2500")
          i = 2500
          exit
       end if
       write(errstr,*) "Current directory : ",trim(adjustl(currdir))
       call message(AppMsg,Noisy)
       currpath_F = trim(adjustl(currdir)) // trim(adjustl(eml_F))
       inquire(FILE=trim(currpath_F), EXIST=exists)
       if (.not. exists) then
          write(errstr,*)trim(adjustl(currpath_F)),' does not exist, disgarding...'
          call warning("noEMLLIST",trim(adjustl(errstr)))
          cycle
       endif
       ! sanity check
       status = 0
       warndie = .TRUE. ! 1 = read
       call check_open_set(trim(adjustl(currpath_F)),inEML_S,1,status,warndie)
       if (status .eq. 1) then
          write(errstr,*)trim(adjustl(currpath_F)),' cannot open or read, disgarding...'
          call warning("badEMLLIST",trim(adjustl(errstr)))
          cycle
       end if
       if (.not.hasBlock(inEML_S,'SRCLIST')) then
          write(errstr,*)trim(adjustl(currpath_F)),' missing SRCLIST block, disgarding...'
          call warning("badEMLLIST",trim(adjustl(errstr)))
          cycle
       else
          j = j + 1
          fullpath(i) = trim(adjustl(currpath_F)) ! accumulated 'good ones'
       endif
       call release(inEML_S)
    enddo
    ! free up some wasted memory
    allocate(emllist(j))
    do i = 1, j
       emllist(i) = fullpath(i)
    end do
    deallocate(fullpath)
    !
    ! open the output merged source list
    !
    write(errstr,*) 'Found ',j,' ok to process srclist files in ',trim(adjustl(outEML_F))
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE. ! 3 = create
    call check_open_set(trim(adjustl(outEML_F)),outEML_S,3,status,warndie)
    !
    if (j .lt. 1) then
       call error("noEMLfound","None of srclist files in dirlist exist or are readable")
    endif
    !
    ! loop through actual good emllist.fits files
    !    
    do i = 1, j
       inEML_F = emllist(i)
       status = 0
       warndie = .TRUE. ! 1 = read
       call check_open_set(trim(adjustl(inEML_F)),inEML_S,1,status,warndie)
       inEML_T = table(inEML_S,'SRCLIST', MODIFY)
       nRows = numberOfRows(inEML_T) ! channels or events, depending
       write(errstr,*)'Current emllist.fits file has ',nRows,' rows.'
       !
       if (.not.hasColumn(inEML_T,'ML_ID_SRC')) then
          call error("noMLIDSRC","Current EML file does not contain an ML_ID_SRC column.")
       else
          mlidsrc_C = column(inEML_T,'ML_ID_SRC', READ)
          mlidsrc_JP => int32Data(mlidsrc_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'ID_INST')) then
          call error("noIDINST","Current EML file does not contain an ID_INST column.")
       else
          idinst_C = column(inEML_T,'ID_INST', READ)
          idinst_JP => int32Data(idinst_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'ID_BAND')) then
          call error("noIDBAND","Current EML file does not contain an ID_BAND column.")
       else
          idband_C = column(inEML_T,'ID_BAND', READ)
          idband_JP => int32Data(idband_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'SCTS')) then
          call error("no","Current EML file does not contain a SCTS column.")
       else
          scts_C = column(inEML_T,'SCTS', READ)
          scts_RP => real32Data(scts_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'DET_ML')) then
          call error("noDETML","Current EML file does not contain a DET_ML column.")
       else
          detml_C = column(inEML_T,'DET_ML', READ)
          detml_RP => real32Data(detml_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'FLUX')) then
          call error("noFLUX","Current EML file does not contain a FLUX column.")
       else
          flux_C = column(inEML_T,'FLUX', READ)
          flux_RP => real32Data(flux_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'FLUX_ERR')) then
          call error("noFLUXERR","Current file does not contain a FLUX_ERR column.")
       else
          fluxerr_C = column(inEML_T,'FLUX_ERR', READ)
          fluxerr_RP => real32Data(fluxerr_C)        
       endif
       !
       if (.not.hasColumn(inEML_T,'RATE')) then
          call error("noRATE","Current EML file does not contain a RATE column.")
       else
          rate_C = column(inEML_T,'RATE', READ)
          rate_RP => real32Data(rate_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'RATE_ERR')) then
          call error("noRATEERR","Current file does not contain a RATE_ERR column.")
       else
          rateerr_C = column(inEML_T,'RATE_ERR', READ)
          rateerr_RP => real32Data(rateerr_C)        
       endif
       !
       if (.not.hasColumn(inEML_T,'RA')) then
          call error("noRA","Current EML file does not contain an RA column.")
       else
          ra_C = column(inEML_T,'RA', READ)
          ra_RP => real32Data(ra_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'DEC')) then
          call error("noDEC","Current EML file does not contain a DEC column.")
       else
          dec_C = column(inEML_T,'DEC', READ)
          dec_RP => real32Data(dec_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'RADEC_ERR')) then
          call error("noRADECERR","Current EML file does not contain an RADEC_ERR column.")
       else
          radecerr_C = column(inEML_T,'RADEC_ERR', READ)
          radecerr_RP => real32Data(radecerr_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'DIST_NN')) then
          call error("noDISTNN","Current EML file does not contain a DIST_NN column.")
       else
          distnn_C = column(inEML_T,'DIST_NN', READ)
          distnn_RP => real32Data(distnn_C) 
       endif
       !



       
    end do
    close(lunout)

    return

  end subroutine emlmerge

end module emlmerge_mod
